# NGAS Utility

import os
import shutil
import requests
import jxmlease
import time
from pycapo import CapoConfig
import random
import uuid
from urllib.parse import quote_plus
import logging

from pyat.ingestion import evlasciencedatamodel as sc

capoprofile = os.environ['CAPO_PROFILE']
config = CapoConfig(profile=capoprofile)
# Retrieve NGAS host information
NGASHOSTS = [ x.strip() for x in config.getstring('archive-ingestion.NGASHosts').split(' ') ]
archivecommand = {'status': 'STATUS', 'archive': 'ARCHIVE', 'retrieve': 'RETRIEVE'}
LOG = logging.getLogger('ingestion')

def random_ngas_host():
    return random.choice(NGASHOSTS)

def ingestevlasdmsintongas(filesetbasepath, bdfpath):
    filesets = []
    for fileset in filesets():
        filesetpath = os.path.join(filesetbasepath, fileset)
        print('Processing fileset: '.format(fileset))


def ingestsdmintongas(sdm, sdmpath, ngas_staging):
    # Copy SDM table files
    for sdmtable in sdm.sdmtables:
        table = sdm.sdmtables[sdmtable]
        if isinstance(sdm.sdmtables[sdmtable], sc.SDMXMLTable):
            fromextension = '.xml'
            toextension = '.sdm'
        elif isinstance(sdm.sdmtables[sdmtable], sc.SDMBinaryTable):
            fromextension = '.bin'
            toextension = '.bin'
        else:
            LOG.critical('Unrecognizable SDM table type')
            exit(-1)

        fromfile = os.path.join(sdmpath, table.tablename + fromextension)
        tofilename = table.entityid.replace(':', '_').replace('/', '_') + toextension
        tofile = os.path.join(ngas_staging, tofilename)
        LOG.info('SDM ingestion copying from {} to staging area {}'.format(fromfile, tofile))
        shutil.copy(fromfile, tofile)
        # filelist.append(tofile)

    # Copy BDF files
    for bdffile in sdm.bdffiles:
        extension = '.bdf'
        filename = bdffile.entityid.replace(':', '_').replace('/', '_')
        # fromfile = os.path.join(sdmpath, filename) + extension
        fromfile = os.path.join(sdmpath, filename)
        tofile = os.path.join(ngas_staging, filename + extension)
        # TODO don't copy bdfs yet - disk space
        # Locally copy bdf files, otherwise link to them
        LOG.info('Copying (or linking) BDF file from {} to {}'.format(fromfile, tofile))
        shutil.copy(fromfile, tofile)
        # os.link(fromfile, tofile)
        # filelist.append(tofile)

    # if len(filelist) > 0:
    #     archivefiles(filelist)
    # else:
    #     logger.warning('No SDM tables or BDF files to process.')


def generate_science_product_locator(telescope, filetype):
    """
    Generate unique science product locator for science products
    :param telescope: the telescope
    :param filetype: the type of the science product
    :return: science product locator
    """

    id = uuid.uuid4()
    science_product_locator = f"uid://{telescope.lower()}/{filetype}/{id}"
    return science_product_locator

def generate_uuids_for_ngas(telescope, filetype, extension):
    """
    Generate unique uuids for ngas storage
    :param filename: the current name of the file
    :return: new file name
    """

    id = uuid.uuid4()
    file = f"uid____{telescope}_{filetype}_{id}.{extension}"
    return file

def archivefiles(archivefilelist):
    filecount = 0
    for archivefile in archivefilelist:
        # archivefile 0:filename 1: file size on disk
        starttime = time.time()
        NGASHOST = random_ngas_host()
        filecount += 1
        ngascmd = NGASHOST + archivecommand['archive'] + '?filename=file://' + quote_plus(archivefile[0])
        LOG.debug(ngascmd)
        response = requests.get(ngascmd)
        returncode = response.status_code
        archivemsg = 'Archive request ({}): {} [{}] in {}s'.format(filecount, ngascmd, returncode, time.time() - starttime) # fix units
        LOG.info(archivemsg)
        doc = jxmlease.parse(response.text)
        if returncode == 200:
            expected_file_size = archivefile[1]
            ngas_file_size = int(doc['NgamsStatus']['DiskStatus']['FileStatus'].get_xml_attr('FileSize'))
            LOG.debug(f"Ingested file size: {ngas_file_size} Expected file size: {expected_file_size}")
            if ngas_file_size != expected_file_size:
                LOG.error(f"NGAS ingested file size {ngas_file_size} does not match expected file size {expected_file_size}")
        else:
            error_message = doc['NgamsStatus']['Status'].get_xml_attr('Message')
            LOG.error(f"NGAS ingestion of {archivefile[0]} failed. Return code: {returncode} Message: {error_message}")


def test_uuid(telescope, filetype, extension):
    print(generate_uuids_for_ngas(telescope, filetype, extension))


if __name__ == '__main__':
    generate_science_product_locator('EVLA', 'execblock')
