import os
import sys
import time
import logging
import re

import numpy as np
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time, TimeDelta
from astropy.coordinates import SkyCoord

from pyat.ingestion import archiveutils as au

LOG = logging.getLogger('ingestion')

"""
Represents a VLBA IDIFITS file
"""
class VLBAFits:
    def __init__(self, fits_file):
        self._idifitsfile = fits_file
        if not os.path.exists(self.getfitsfile()) or os.path.getsize(self.getfitsfile()) == 0:
            LOG.error(f"IDIFITS file {self.getfitsfile()} could not be found or is empty.")
            exit(-1)
        self._scans = self._build_scans(self.getfitsfile())

    def _build_scans(self, fitsfile):
        """
        Build a table of scans from the IDIFITS file. A scan occurs when the source id changes between
        rows and, in the case of multiple UV_DATA tables, when a new table is encountered (assumption
        that is consistent with the legacy archive).
        """
        scans_start_time = time.time()
        with fits.open(fitsfile) as hdu:
            self._project = self._getproject(hdu)
            self._correlator_pass = self._getcorrelatorpass()
            self._segment = self._getsegment(hdu)
            table = 0
            scan = 0
            number_rows = 0
            scan_table_rows = []
            while True:
                # Extracting the UV_DATA table(s) in this way is done because older IDIFITS files
                # were permitted to have multiple UV_DATA tables and the is no way to extract the
                # list of those tables. hdu['UV_DATA'] only returns the first table
                try:
                    if 'EXTNAME' in hdu[table].header:
                        if hdu[table].name == 'UV_DATA':
                            # New UV_DATA table triggers a new scan. This is consistent with the legacy
                            # archive, but seems suspicious. Good question for VLBA subsystem scientist.
                            scan = scan + 1 # New table triggers new scan (??)
                            uv_data = hdu[table].data
                            uv_hdr = hdu[table].header

                            polarizations = self.getpolarizations(hdu['UV_DATA'].header)

                            uv_length = uv_hdr['NAXIS2']
                            if uv_length == 0:
                                LOG.error(f"No UV_DATA table for file {self.getfitsfile()}")
                                exit(-1)
                            scan_start_row = 0
                            for row in range(uv_length):
                                number_rows = number_rows + 1

                                if (row < uv_length - 1) and (uv_data[row]['SOURCE'] != uv_data[row+1]['SOURCE']):
                                    # Scan break
                                    src_mask = self.getsourcetable()[self._source_id_column] == uv_data[row]['SOURCE']
                                    source, source_ra, source_dec = self._getsourceinfo(src_mask)
                                    frequency_id = uv_data[row]['FREQID']
                                    integration_time = uv_data[row]['INTTIM']
                                    frequencies = self.getfrequencies(frequency_id)
                                    bandwidth = self.getbandwidths(frequency_id)
                                    start_time = Time(uv_data[scan_start_row]['DATE'], format='jd') + TimeDelta(uv_data[scan_start_row]['Time'], format='jd')
                                    stop_time = Time(uv_data[row]['DATE'], format='jd') + TimeDelta(uv_data[row]['Time'], format='jd')

                                    scan_table_rows.append((self.getproject(), scan, source, start_time.mjd, stop_time.mjd, round((stop_time - start_time).sec, 2), integration_time, ', '.join(str(x) for x in frequencies), bandwidth, polarizations, source_ra, source_dec))

                                    scan_start_row = row + 1
                                    scan = scan + 1
                                elif (row == uv_length - 1) and (uv_data[row]['SOURCE'] != uv_data[row-1]['SOURCE']):
                                    # Last row in uv_data table and different source as next to last row
                                    src_mask = self.getsourcetable()[self._source_id_column] == uv_data[row]['SOURCE']
                                    source, source_ra, source_dec = self._getsourceinfo(src_mask)
                                    frequency_id = uv_data[row]['FREQID']
                                    integration_time = uv_data[row]['INTTIM']
                                    frequencies = self.getfrequencies(frequency_id)
                                    bandwidth = self.getbandwidths(frequency_id)
                                    start_time = Time(uv_data[row]['DATE'], format='jd') + TimeDelta(uv_data[row]['Time'], format='jd')
                                    stop_time = Time(uv_data[row]['DATE'], format='jd') + TimeDelta(uv_data[row]['Time'], format='jd')

                                    scan_table_rows.append((self.getproject(), scan, source, start_time.mdd, stop_time.mjd, round((stop_time - start_time).sec, 2), integration_time, ', '.join(str(x) for x in frequencies), bandwidth, polarizations, source_ra, source_dec))
                                elif row == uv_length - 1:
                                    # Last row in uv_data table and same source as next to last row
                                    src_mask = self.getsourcetable()[self._source_id_column] == uv_data[row]['SOURCE']
                                    source, source_ra, source_dec = self._getsourceinfo(src_mask)
                                    frequency_id = uv_data[row]['FREQID']
                                    integration_time = uv_data[row]['INTTIM']
                                    frequencies = self.getfrequencies(frequency_id)
                                    bandwidth = self.getbandwidths(frequency_id)
                                    start_time = Time(uv_data[scan_start_row]['DATE'], format='jd') + TimeDelta(uv_data[scan_start_row]['Time'], format='jd')
                                    stop_time = Time(uv_data[row]['DATE'], format='jd') + TimeDelta(uv_data[row]['Time'], format='jd')

                                    scan_table_rows.append((self.getproject(), scan, source, start_time.mjd, stop_time.mjd, round((stop_time - start_time).sec, 2), integration_time, ', '.join(str(x) for x in frequencies), bandwidth, polarizations, source_ra, source_dec))
                        elif hdu[table].header['EXTNAME'] == 'SOURCE':
                            # Construct the source table
                            src_table = Table(hdu[table].data)
                            if 'ID_NO.' in src_table.columns:
                                self._source_id_column = 'ID_NO.'
                            elif 'SOURCE_ID' in src_table.columns:
                                self._source_id_column = 'SOURCE_ID'

                            self._source_table = src_table
                        elif hdu[table].header['EXTNAME'] == 'FREQUENCY':
                            # Construct the frequency table and get the header
                            frequency_header = hdu[table].header
                            frequency_table = Table(hdu[table].data)
                            self._ref_frequency = frequency_header['REF_FREQ']
                            self._frequency_table = frequency_table

                    table = table + 1
                except OSError:
                    break
                except IndexError:
                    break

        # The column names are the interface to persistVLBAmetadata. Be careful!
        # The order in scan_table_rows must correspond to the order of the columns names below. Be careful!
        scan_table_column_names = ('Project', 'Scan', 'SOURCE', 'START_TIME', 'END_TIME',
                                   'TIME_ON_SOURCE', 'INTEGRATION_TIME', 'FREQUENCY',
                                   'BANDWIDTH', 'POLARIZATION', 'RA', 'DEC')
        scan_table = Table(rows=scan_table_rows, names=scan_table_column_names)
        LOG.info(f"File {self.getfitsfile()} has {number_rows} rows and {scan} scans.")
        LOG.info(f"{scan} scans built in {time.time() - scans_start_time}s")
        return scan_table # Return a table of scans where each row is built by build_scan

    def getfitsfile(self):
        return self._idifitsfile

    def getscantable(self):
        return self._scans

    def _getsourceinfo(self, src_mask):
        source = self.getsourcetable()[src_mask]['SOURCE'][0]
        source_ra = self.getsourcetable()[src_mask]['RAAPP'][0]
        source_dec = self.getsourcetable()[src_mask]['DECAPP'][0]

        c = SkyCoord(source_ra, source_dec, frame='icrs', unit='deg')
        # For debugging
        source_radec = c.to_string('hmsdms')

        # Converting to radians
        source_ra_rad = c.ra.radian
        source_dec_rad = c.dec.radian

        # Currently returning the ra and dec in degrees, but modifying the return allows
        # using radians or an 'hmsdms' string
        return source, source_ra, source_dec

    def _getproject(self, hdu):
        """
        Extract the project code from the observer string in the FITS file
        e.g., BB100 from BB100A1
        """
        m = re.match('([A-Z]+\d+)(.*)', hdu['PRIMARY'].header['OBSERVER'])
        return m.group(1)

    def getproject(self):
        return self._project

    def _getsegment(self, hdu):
        """
        Extract the segment code from the observer string in the FITS file
        e.g., A1 from BB100A1
        """
        m = re.match('([A-Z]+\d+)(.*)', hdu['PRIMARY'].header['OBSERVER'])
        return m.group(2)

    def getsegment(self):
        return self._segment

    def _getcorrelatorpass(self):
        """ Extract the correlator pass from the file name (not present in the IDIFITS file) """
        fflist = os.path.basename(self.getfitsfile()).split('_')
        return fflist[2]

    def getcorrelatorpass(self):
        return self._correlator_pass

    def getsourcetable(self):
        return self._source_table

    def getreffrequency(self):
        return self._ref_frequency

    def getfrequencytable(self):
        return self._frequency_table

    def getfrequencies(self, frequency_id):
        frequency_row = self.getfrequencytable()[self.getfrequencytable()['FREQID'] == frequency_id]
        frequencies = frequency_row[0]['BANDFREQ']
        if np.isscalar(frequencies):
            # If the frequency is a scalar, put it in a numpy ndarray
            freq_array = np.ndarray(shape=(1))
            freq_array[0] = frequency_row[0]['BANDFREQ']
            return freq_array
        return (self.getreffrequency() + frequencies) / 1000000

    def getbandwidths(self, frequency_id):
        frequency_row = self.getfrequencytable()[self.getfrequencytable()['FREQID'] == frequency_id]
        # If the frequency is a scalar, put it in a numpy ndarray
        if np.isscalar(frequency_row[0]['TOTAL_BANDWIDTH']):
            return frequency_row[0]['TOTAL_BANDWIDTH'] / 1000000
        else:
            return frequency_row[0]['TOTAL_BANDWIDTH'][0] / 1000000

    def getpolarizations(self, uv_header):
        numpolarizations = uv_header['NO_STKD']
        polarization_code = 0
        for i in range(numpolarizations):
            pol_key = f"STK_{i+1}"
            if pol_key not in uv_header.keys():
                LOG.warning(f"Stokes key: {pol_key} is not present in UV_DATA table")
                continue
            polarization = self.lookup_polarization(uv_header[pol_key])
            polarization_code = polarization_code + polarization
        return au.getpolarization(polarization_code)

    def lookup_polarization(self, polarization_code):
        """ Return the polarization code for the numeric polarization in the IDIFITS file """
        polarization_codes = {
            1:  'I',
            2:  'Q',
            3:  'U',
            4:  'V',
            -1: 'RR',
            -2: 'LL',
            -3: 'RL',
            -4: 'LR',
            -5: 'XX',
            -6: 'YY',
            -7: 'XY',
            -8: 'YX'
        }

        if polarization_code not in polarization_codes.keys():
            LOG.warn(f"Invalid polarization code: {polarization_code}")
            exit(-1)

        return au.getpolarizationcode(polarization_codes[polarization_code])

    def lookup_observing_band(self):
        """ Lookup the observing band from the reference frequency """
        ref_freq = self.getreffrequency() / 1000000000
        obs_band = ''

        if 0.058 <= ref_freq <= 0.084:
            obs_band = '4'
        elif 0.236 <= ref_freq <= 0.492:
            obs_band = 'P'
        elif 1.0 <= ref_freq <= 2.0:
            obs_band = 'L'
        elif 2.0 < ref_freq <= 4.0:
            obs_band = 'S'
        elif 4.0 < ref_freq <= 8.0:
            obs_band = 'C'
        elif 8.0 < ref_freq <= 12.0:
            obs_band = 'X'
        elif 12.0 < ref_freq <= 18.0:
            obs_band = 'Ku'
        elif 18.0 < ref_freq <= 26.5:
            obs_band = 'K'
        elif 26.5 < ref_freq <= 40.0:
            obs_band = 'Ka'
        elif 40.0 < ref_freq <= 50.0:
            obs_band = 'Q'
        else:
            LOG.warning(f"Reference frequency {ref_freq} could not be located in a band")

        return obs_band


if __name__ == '__main__':
    if (len(sys.argv) != 3):
        LOG.error(f"Usage: VLBAIDIFits path file")
        exit(-1)

    path = sys.argv[1]
    file = sys.argv[2]

    fits = VLBAFits(os.path.join(path, file))
    fits.getscantable().pprint(max_lines=250, max_width=250)


