import datetime
import logging
import os
import tarfile
import hashlib
from pathlib import Path

from pycapo import CapoConfig
import pyat.ingestion.archiveutils as au
import pyat.ingestion.almautils as alu
from pyat.ingestion import NGASUtil as ng
from pyat.schema import create_session
from pyat.schema.model import Filegroup, File, Calibration, ScienceProduct, ProductGroup

LOG = logging.getLogger('ingestion')

session = create_session('SDM')

config = CapoConfig()
site_name = config.getstring('archive-ingestion.NGASSite')
ngas_cluster = config.getstring('archive-ingestion.NGASCluster')


def ingestcaltable(file):
    """
    Ingest a single calibration product tar file. Create the metadata and
    store the file on NGAS
    :param file: the tar file to be ingested
    :return:
    """
    # No filegroup is given, so it must be extracted from the tar file
    found = False
    tar = tarfile.open(file)
    for tarinfo in tar:
        # ms.calapply.txt has the file set id embedded in it
        if tarinfo.isreg() and tarinfo.name.endswith('.ms.calapply.txt'):
            found = True
        if found:
            fsstart = tarinfo.name.find('products/')
            fsend = tarinfo.name.find('.ms')
            fileset = tarinfo.name[fsstart + 9:fsend]
            ingestcalmetadata(file, fileset)
            break
    if not found:
        LOG.error('Unable to find matching file group for calibration products file {}'.format(file))


def ingestcalmetadata(file, external_name, nongas):
    """
    Ingest a single calibration product tar file, given the file and associated file group
    :param file: the tar file to be ingested
    :param external_name: the external name of the parent execution block
    :param nongas: flag to indicate if ingestion to NGAS should not occur
    :return:
    """

    ngas_id = ng.generate_uuids_for_ngas('evla', 'calibration', 'tar')
    exec_block = au.get_execution_block_from_external_name(external_name)
    LOG.debug(f"EB ID: {exec_block.ngas_fileset_id}")
    if exec_block is None:
        LOG.error(f"Parent execution block {external_name} not found for calibration products {file}")
        exit(-2)
    else:
        if exec_block.filegroup is None:
            LOG.error(f"Execution block filegroup not found for execution block {external_name}")
            exit(-2)

        calibration_filegroup = au.get_filegroup_from_name(file.name)

        if calibration_filegroup is not None:
            # Calibration products file exists, replace
            calfile = au.get_file_from_group(calibration_filegroup)
            calfile.filesize = file.stat().st_size,
            calfile.checksum=sha256(file)
            calfile.checksum_type='SHA256'
            calfile.ingestion_time=datetime.datetime.utcnow()
            calfile.ngas_id=ngas_id
            LOG.info(f"Replacing calibrations products: {file.name} for execution block: {external_name}")
        else:
            calibration_filegroup = Filegroup(parent_filegroup=exec_block.filegroup, type='calibration')
            session.add(calibration_filegroup)
            filesize = file.stat().st_size
            site_name = config.getstring('archive-ingestion.NGASSite')
            ngas_cluster = config.getstring('archive-ingestion.NGASCluster')
            calfile = File(file_path=groupname,
                           ngas_id=ngas_id,
                           ngas_location=site_name,
                           ngas_cluster=ngas_cluster,
                           filename=os.path.basename(file),
                           filegroup1=calibration_filegroup,
                           filesize=filesize,
                           format='tar',
                           type='cal',
                           checksum=sha256(file),
                           checksum_type='SHA256',
                           ingestion_time=datetime.datetime.utcnow())

            calibration_filegroup.datasize = calfile.filesize
            session.add(calfile)

            science_product_locator = ng.generate_science_product_locator(execblock.telescope.lower(), 'calibration')

            calibration = Calibration(
                execution_block=execblock,
                alma_ous_id=None,
                filegroup=calibration_filegroup,
                project_code=exec_block.project_code,
                science_product_locator=science_product_locator
            )

            science_product = ScienceProduct(
                science_product_locator=science_product_locator,
                science_product_type='Calibration',
                metadata_ingestion_date=datetime.datetime.utcnow(),
                metadata_ingestion_version=1, # TBD
                filegroup=calibration_filegroup
            )

            # Ordering matters to satisfy FK constraints:
            session.add(science_product)
            session.add(calibration)

            # Add to science_products_projects
            project = execblock.project
            project.science_products.append(science_product)
            session.add(project)

        session.commit()

        # Now do the NGAS ingestion, if not disabled
        if nongas:
            LOG.info(f"NGAS ingestion is disabled. File {ngas_id} will not be ingested.")
        else:
            LOG.info(f"Linking file {file} to {str(Path(file.parent) / ngas_id)}")
            os.link(file, file.parent / ngas_id)
            ingestcalfile(file.parent / ngas_id, calfile.filesize)


def ingestalmacalibration(mous):
    LOG.debug(f"MOUS for calibrations: {mous}")
    # Need the eb filegroup
    # Get eb from MOUS
    projectcode = au.getprojectcodefrommous(mous)
    LOG.debug(f"Project code: {projectcode}")
    calfilegroup = Filegroup(groupname=mous, project_code=projectcode, type='calibration')
    session.add(calfilegroup)

    # Now get the files and create File entries
    # Get the list of files - Alma query
    calfiles = alu.getalmacalfiles(mous)
    LOG.info(f"Cal files: {str(calfiles)}")
    # ALMA calibration files are not stored on our NGAS system, so indicate that:
    for file in calfiles:
        calfile = File(
            ngas_id = file[0],
            ngas_location='NAASC',
            ngas_cluster='NAASC',
            filename=file[0],
            filesize=file[1],
            filegroup1=calfilegroup,
            format='cal',
            type='cal'
        )
        session.add(calfile)

    # Now do the calibrations table
    calibrations = Calibration(
        alma_ous_id=mous,
        filegroup=calfilegroup,
        project_code=projectcode
    )
    session.add(calibrations)
    # session.commit()


def ingest_from_manifest(manifest):
    telescope = manifest.get_telescope()
    path, calibration_tar_file, input_product_locators, reingest, ngas_ingest = manifest.get_calibration_ingestion_parameters()
    ngas_id = ng.generate_uuids_for_ngas('evla', 'calibration', 'tar')
    # In theory, I should be able to pass a list of execution block science product locators
    ingest_calibration(telescope, input_product_locators, path, calibration_tar_file, ngas_id, reingest)

    if ngas_ingest:
        LOG.info(f"TEST Linking file {os.path.join(path, calibration_tar_file)} to {os.path.join(path, ngas_id)}")
        # LOG.info(f"Linking file {os.path.join(path, calibration_tar_file)} to {os.path.join(path, ngas_id)}")
        # os.link(os.path.join(path, calibration_tar_file), os.path.join(os.path.dirname(file), ngas_id))
        # ingestcalfile(os.path.join(path, ngas_id), os.path.getsize(os.path.join(path, ngas_id)))
    else:
        LOG.info(f"NGAS ingestion is disabled. File {calibration_tar_file} will not be ingested.")


def xingest_from_manifest(manifest):
    telescope = manifest.get_telescope()
    # Pass telescope, path to
    path, calibration_tar_file, input_product_locators, reingest, ngas_ingest = manifest.get_calibration_ingestion_parameters()
    ngas_id = ng.generate_uuids_for_ngas('evla', 'calibration', 'tar')
    # In theory, I should be able to pass a list of execution block science product locators
    ingest_calibration(telescope, input_product_locators, path, calibration_tar_file, ngas_id, reingest)

    if ngas_ingest:
        LOG.info(f"TEST Linking file {os.path.join(path, calibration_tar_file)} to {os.path.join(path, ngas_id)}")
        # LOG.info(f"Linking file {os.path.join(path, calibration_tar_file)} to {os.path.join(path, ngas_id)}")
        # os.link(os.path.join(path, calibration_tar_file), os.path.join(os.path.dirname(file), ngas_id))
        # ingestcalfile(os.path.join(path, ngas_id), os.path.getsize(os.path.join(path, ngas_id)))
    else:
        LOG.info(f"NGAS ingestion is disabled. File {calibration_tar_file} will not be ingested.")


def ingest_calibration(telescope, execution_block_science_product_locators, path, calibration_tar_file, ngas_id, reingest):
    """
    Ingest calibration products, attach to executiion block, and add to product groups
    :param telescope: telescope (evla or alma)
    :param execution_block_science_product_locator: list of science product locators of the execution blocks that were calibrated
    :param path: Location of calibration products file
    :param calibration_tar_file: tar file containing the calibration products
    :param reingest: should existing products be delete and reingested?
    :return:
    """

    # Get the parent eb using its science product locator; for now, only do the first
    # since there should only be one
    execblock = au.get_execution_block(execution_block_science_product_locators[0])
    if execblock is None:
        LOG.error(f"Execution block with science product locator {execution_block_science_product_locator} not found.")
        exit(-1)
    execution_block_science_product = au.get_science_product_by_spl(execution_block_science_product_locator)

    # Add the calibration products filegroup and file
    # Look for calibration by filename, in case already ingested
    calibration_products_file = au.get_file(calibration_tar_file)
    if calibration_products_file is not None:
        if reingest:
            # TODO fix reingestion
            LOG.debug(f"Reingestion of cal products requested for {calibration_tar_file}")
            # Delete the existing entry
        else:
            LOG.error(f"Calibration products file {calibration_tar_file} already ingested")
            exit(-2)
    else:
        LOG.debug(f"Initial ingestion of cal products requested for {os.path.join(path, calibration_tar_file)}")

    calibration_filegroup = Filegroup(
        groupname=calibration_tar_file,
        parent_filegroup=execblock.filegroup,
        type='calibration'
    )
    db_session = session.object_session(calibration_filegroup)
    db_session.add(calibration_filegroup)

    calibration_products_file = File(
        file_path=os.path.dirname(path),
        ngas_id=ngas_id,
        ngas_location=site_name,
        ngas_cluster=ngas_cluster,
        filename=calibration_tar_file,
        filegroup1=calibration_filegroup,
        filesize=os.path.getsize(os.path.join(path, calibration_tar_file)),
        format='tar',
        type='cal',
        checksum=sha256(os.path.join(path, calibration_tar_file)),
        checksum_type='SHA256',
        ingestion_time=datetime.datetime.utcnow()
    )
    db_session.add(calibration_products_file)

    science_product_locator = ng.generate_science_product_locator(execblock.telescope.lower(), 'calibration')

    calibration = Calibration(
        project_code=execblock.project_code,
        execution_block=execblock,
        alma_ous_id=None,
        filegroup=calibration_filegroup,
        science_product_locator=science_product_locator
    )

    db_session.add(calibration)

    science_product = ScienceProduct(
        science_product_locator=science_product_locator,
        science_product_type='Calibration',
        metadata_ingestion_date=datetime.datetime.utcnow(),
        metadata_ingestion_version=1,
        filegroup=calibration_filegroup
    )

    db_session.add(science_product)

    # TODO ancillary products for calibrations

    # Add to science_products_projects
    execblock.project.science_products.append(science_product)

    # Add to product_groups and science_products_product_groups
    input_product_group = ProductGroup(
        product_group_type='input_group',
        parent_product_group=None,
        ALMA_ous=None
    )
    input_product_group.science_products.append(execution_block_science_product)
    db_session.add(input_product_group)

    output_product_group = ProductGroup(
        product_group_type='output_group',
        parent_product_group=input_product_group,
        ALMA_ous=None
    )
    output_product_group.science_products.append(science_product)
    db_session.add(output_product_group)

    db_session.commit()


# def ingest_calibrationx(telescope, execution_block_science_product_locators, path,
#                         calibration_tar_file, ngas_id, reingest):


def ingestcalfile(calibration_file, file_size):
    ingestion_list = []
    ingestion_list.append([calibration_file, file_size])
    LOG.info(f"Ingesting file: {calibration_file.name}")
    ng.archivefiles(ingestion_list)


def sha256(fname):
    hash_sha256 = hashlib.sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


def test1():
    LOG.debug("TEST")


if __name__ == '__main__':
    test1()

